function createAsyncMutex() {
	let current_promise = null;

	return {
		async acquire() {
			if (current_promise !== null) {
				await current_promise;
			}

			let unlock;

			current_promise = new Promise(resolve => {
				unlock = resolve;
			});

			return () => {
				unlock();
				current_promise = null;
			}
		}
	}
}

async function handleRequest(state, message) {
	const request_handler = state.request_handler_fn;
	let response = null, error = null;

	try {
		response = await request_handler(message.body);
	} catch (_) {
		error = _;
	}

	if (error === null) {
		state.sendRawMessage({
			cmd: "response",
			request_id: message.request_id,
			success: true,
			body: response
		});
	} else {
		state.sendRawMessage({
			cmd: "response",
			request_id: message.request_id,
			success: false,
			error: error.message
		});
	}
}

async function onMessage(state, str) {
	const message = JSON.parse(str);

	if (message.cmd === "init") {
		state.sendRawMessage({
			ready: message.ready_token
		});
	} else if (message.cmd === "request") {
		const mutex = state.request_handler_mutex;

		const release = await mutex.acquire();

		await handleRequest(state, message);

		release();
	}
}

function index(send_message_fn, request_handler_fn) {
	let state = {
		sendRawMessage(data) {
			send_message_fn(JSON.stringify(data) + "\n");
		},
		request_handler_mutex: createAsyncMutex(),
		request_handler_fn
	};

	return {
		onMessage(...args) {
			return onMessage(state, ...args)
		}
	}
}

function init(requestHandler) {
	const slave = index(this.sendMessage, requestHandler);

	this.onMessage = slave.onMessage;
}

async function WebWorkerMain(request_handler_url) {
	let {requestHandler} = await import(request_handler_url);

	init.call(this, requestHandler);
}

async function WorkerMain(request_handler_path) {
	let {requestHandler} = await import(request_handler_path);

	init.call(this, requestHandler);
}

export { WebWorkerMain, WorkerMain };
